#include "dwt.h"

void dwt::readfilter(char* name) {
  //hvis filter.name eksisterer
  char buf[80];
  ifstream file(name);
  if(file==NULL) { cout << "file not existing" << endl; }
  file.getline(buf,80);
  while(*buf == '#') {
    file.getline(buf,80);
  }
  filterlength=atoi(buf); 
  //lav det ene filter
  for(int i=0;i<filterlength;i++) {
    file.getline(buf,80);
    filterh[i]=strtod(buf,NULL);
  }
  file.close();

  //lav de andre filtre ud fra det foerste
  for(int i=0;i<filterlength;i++) {
    if(i%2==0) { 
       filterg[i]=filterh[filterlength-i-1]; 
    }
    else{
      filterg[i]=-filterh[filterlength-i-1]; 
    }
  }
  for(int i=0;i<filterlength/2;i++) {
    filterhinv[i]=filterh[filterlength-2-2*i];
    filterhinv[i+filterlength/2]=filterg[filterlength-2-2*i];
    filterginv[i]=filterh[filterlength-1-2*i];
    filterginv[i+filterlength/2]=filterg[filterlength-1-2*i];
  }
}

void dwt::transform(int antal) {
  (*this).crop(antal);
  no_of_trans = antal;
  int j = rows;
  for(int i=0;i<antal;i++) {
    (*this).transcolumns(j);
    j = j/2;
  } 
  j = columns;
  for(int i=0;i<antal;i++) {
    (*this).transrows(j);
    j = j/2;
  } 
}

void dwt::transrows(int c) {
  double TMP1[c];
  for(int i=0;i<rows;i++) {
    for(int j=0;j<c;j++) {
      TMP1[j]=A[i*columns+j];
    }
    trans1dim(TMP1,c);
    for(int j=0;j<c;j++) {
      A[i*columns+j]=TMP1[j];
    }
  }
}

void dwt::transcolumns(int r) {
  double TMP1[r];
  for(int i=0;i<columns;i++) {
    for(int j=0;j<r;j++) {
      TMP1[j]=A[j*columns+i];
    }
    trans1dim(TMP1,r);
    for(int j=0;j<r;j++) {
      A[j*columns+i]=TMP1[j];
    }
  }
}

void dwt::itrans() {
  int antal = no_of_trans;
  int j = columns;
  for(int i=0;i<antal;i++) {
    j = j/2;
  }
 
  for(int i=0;i<antal;i++) {
    j = 2*j;
    (*this).itransrows(j);
  } 
  j = rows;
  for(int i=0;i<antal;i++) {
    j = j/2;
  }
  for(int i=0;i<antal;i++) {
    j = 2*j;
    (*this).itranscolumns(j);
  } 
}

void dwt::itransrows(int c) {
  double TMP1[c];
  for(int i=0;i<rows;i++) {
    for(int j=0;j<c;j++) {
      TMP1[j]=A[i*columns+j];
    }
    itrans1dim(TMP1,c);
    for(int j=0;j<c;j++) {
      A[i*columns+j]=TMP1[j];
    }
  }
}

void dwt::itranscolumns(int r) {
  double TMP1[r];
  for(int i=0;i<columns;i++) {
    for(int j=0;j<r;j++) {
      TMP1[j]=A[j*columns+i];
    }
    itrans1dim(TMP1,r);
    for(int j=0;j<r;j++) {
      A[j*columns+i]=TMP1[j];
    }
  }
}

// compare is the same as < 
int compare(const void *a,const void *b) {
  double c;
  c = *((double*)a) - *((double*)b);
  if(c<0) {
    return -1;
  }
  if(c>0) {
    return 1;
  }
  return 0;
}

//e can be between 0 and 1
void dwt::compress(double e) {
  double smallest;
  double TMP[columns*rows];
  for(int i=0;i<columns*rows;i++) {
    TMP[i]=fabs(A[i]);
  } 
  qsort(TMP,columns*rows,sizeof(double),compare);
  
  smallest = TMP[(int) rint(e*columns*rows)];
  for(int i=0;i<rows*columns;i++) {
    if(fabs(A[i])<=smallest) {
      A[i]=0;
    }
    else {
      A[i]=rint(A[i]);
    }
  }
}

void dwt::bzwritetrans(char* name,char* format) {
  BZFILE* bzfile;
  int     bzerror;
  int     nWritten;
  int     nBuf;
  char    buf[80];
  int entry=0;
  bzfile = BZ2_bzopen (name,"w" );
  sprintf(buf,"%s\n",format);
  BZ2_bzWrite ( &bzerror, bzfile, buf, strlen(buf) );
  sprintf(buf,"%d %d\n",columns,rows);
  BZ2_bzWrite ( &bzerror, bzfile, buf, strlen(buf) );
  sprintf(buf,"%d\n",no_of_trans);
  BZ2_bzWrite ( &bzerror, bzfile, buf, strlen(buf) );
  for(int i=0;i<columns*rows;i++) {
    if(A[i]!=0) {
      sprintf(buf,"%d %d\n",i-entry,(int) A[i]);
      BZ2_bzWrite ( &bzerror, bzfile, buf, strlen(buf) );
      entry=i;
    }
  }
  BZ2_bzclose(bzfile);
}

char* bzgets(BZFILE* f,char* b,int l) {
  char* q;
  q=b;
  int bzerror;
  for(int i=0;i<l;i++) {
    BZ2_bzRead ( &bzerror, f, q, 1 );
    if(bzerror==BZ_STREAM_END) {
      return NULL;
    }
    if(*q=='\n') {
      q++;
      *q='\0';
      return b;
    }
    else {
      q++;
    }
  }
  return b;
}

void dwt::bzreadtrans(char* name,char* format) {
  int bzerror;
  BZFILE* bzfile = BZ2_bzopen(name,"r");
  char buf[80];
  char *p,*q;
  int i = 0;
  int mod = 1;
  int entry=0;
  delete A;  
  bzgets(bzfile,buf,80);
  if (*buf == *format); //ignore
  bzgets(bzfile,buf,80);
  // find columns
  q=buf;
  for(p=q;!isdigit(*p);p++);
  for(q=p;isdigit(*q);q++);
  *q='\0';
  q++;
  columns = atoi(p);
  // find rows
  for(p=q;!isdigit(*p);p++);
  for(q=p;isdigit(*q);q++);
  *q='\0';
  q++;
  rows =  atoi(p);
  // now make the trans A from the file
  A = new double[columns*rows];
  bzgets(bzfile,buf,80);
  no_of_trans=atoi(buf);
  while(bzgets(bzfile,buf,80)!=Z_NULL) {
    //lav matricen
    q=buf;
    for(p=q;!isdigit(*p);p++);
    for(q=p;isdigit(*q);q++);
    *q='\0';
    q++;
    entry = entry+atoi(p);

    for(p=q;!isdigit(*p) && *p!='-';p++);
    for(q=p;isdigit(*q)||*q=='.'||*q=='-'||*q=='e';q++);
    *q='\0';
    q++;
    A[entry]=strtod(p,NULL);
  }
  BZ2_bzclose(bzfile);
}

void dwt::trans1dim(double* X,int length) {
  double TMP[length];
  for(int i=0;i<length;i++) {
    TMP[i]=0;
  }
  int j= length/2;
  int j1[filterlength];
  for(int i=0;i<j;i++) {
    //udtrk de entries der skal bruges
    j1[0]=2*i;
    for(int k=1;k<filterlength;k++) {
      j1[k]=(j1[k-1]+1)%length;
    }
    //udregn selve transformationen
    for(int k=0;k<filterlength;k++) {
      TMP[i]=TMP[i]+filterh[k]*X[j1[k]];
      TMP[i+j]=TMP[i+j]+filterg[k]*X[j1[k]];
    }
  }
  for(int i=0;i<length;i++) {
    X[i] = TMP[i];
  }
}


void dwt::itrans1dim(double* X,int length) {
  double TMP[length];
  for(int i=0;i<length;i++) {
    TMP[i]=0;
  }
  int j= length/2;
  int j1[filterlength];
  for(int i=0;i<filterlength;i++) {
    j1[i]=1;
  }
  for(int i=0;i<j;i++) { 
    j1[0]=(j+i-filterlength/2+1)%j; //filteret skal have mindst laengde 2
    j1[filterlength/2]=j1[0]+j;
    for(int k=1;k<filterlength/2;k++) { //filterlength skal vaere lige
      j1[k]=(j1[k-1]+1)%j;
      j1[k+filterlength/2]=j1[k]+j;
    }
    //udregn selve transformationen
    for(int k=0;k<filterlength;k++) {
      TMP[2*i]=TMP[2*i]+filterhinv[k]*X[j1[k]];
      TMP[2*i+1]=TMP[2*i+1]+filterginv[k]*X[j1[k]];
    }
  }
  for(int i=0;i<length;i++) {
    X[i] = TMP[i];
  }
}
