#include "matrix.h"

matrix::matrix(int c,int r) {
  columns = c;
  rows = r;
  A = new double[c*r];
  for (int i =0 ; i< c*r ; i++) { A[i]=0; }
}

matrix::matrix() {
  columns = 1;
  rows = 1;
  A = new double[1];
  A[0]=0;
}

double matrix::get(int a,int b ) {
  return A[a*columns + b]; 
}

void matrix::set(int a, int b,double c) {
  A[a*columns + b] = c;
}

matrix* matrix::mult(matrix X,matrix Y) {
  if ( X.columns != Y.rows || X.rows != (*this).rows 
        || (*this).columns != Y.columns ) {
    cout << "incompatible dimensions" << endl;
    return NULL;
  }
  for (int i = 0 ; i < X.rows ; i++) {
    for (int j = 0; j < Y.columns ; j++) {
      (*this).set(i,j,0);
      for ( int k = 0 ; k < X.columns ; k++) {
        (*this).set(i,j,(*this).get(i,j) + X.get(i,k)*Y.get(k,j));
      }
    }
  }
  return this;
}

double matrix::max() {
  double biggest=A[0];
  for (int i=1;i<columns*rows;i++) {
    if(A[i] > biggest) {
      biggest = A[i];
    }
  }
  return biggest;
}

double matrix::min() {
  double smallest=A[0];
  for (int i=1;i<columns*rows;i++) {
    if(A[i] < smallest) {
      smallest = A[i];
    }
  }
  return smallest;
}

void matrix::writepgm(char* name) {
  int j=1;
  int k;
  ofstream file(name);
  file << "P2" << endl; 
  file << columns << " " << rows << endl;
  file << rint(max()) << endl;  
  for(int i=0;i<columns*rows;i++) {
    if(j == 18) { //hver optager hoejst 4 tegn, der maa vaere 70 tegn pr linie
      j = 1; 
      file << endl;
    } 
    k = abs((int) rint(A[i]));
    if(k%10==k) { 
      file << "  " << k << " ";
    }
    else {
      if(k%100==k) {
        file << " " << k << " ";

      }
      else {
        file <<  k << " ";
      }
    }
    j++;
  }
  file.close();
}

void matrix::readpgm(char* name) {
  char buf[80];
  char *p,*q;
  int i = 0;
  int mod = 1;
  delete A;  
  ifstream file(name);
  file.getline(buf,80);
  if (*buf == 'P'); //ignore
  file.getline(buf,80);
  while(*buf == '#') {
    file.getline(buf,80);
  }  

  // find columns
  q=buf;
  for(p=q;!isdigit(*p);p++);
  for(q=p;isdigit(*q);q++);
  *q='\0';
  q++;
  columns = atoi(p);

  // find rows
  for(p=q;!isdigit(*p);p++);
  for(q=p;isdigit(*q);q++);
  *q='\0';
  q++;
  rows =  atoi(p);

  // now make the matrix A from the file
  A = new double[columns*rows];
  file.getline(buf,80); //ignore max value 
  while(file.getline(buf,80)) {
    if(*buf == '#') {
      continue;
    }
    q=buf;
    while(*q!='\0' && q<buf+80 && i < columns*rows) {
     for(p=q;!isdigit(*p);p++);
     for(q=p;isdigit(*q);q++);
     *q='\0';
     q++;
     A[i]=atoi(p);
     i++;
    }    
  }
  file.close();
} 

void matrix::crop(int antal) {
  int mod = 1;
  for(int j = 0;j<antal;j++) {
    mod = mod*2;
  }

  int newrows = rows - rows%mod;
  int newcols = columns - columns%mod;

  double *newA =  new double[newcols*newrows];
  
  for(int k=0;k<newcols;k++) {
    for(int j=0;j<newrows;j++) {
      newA[k + j*newcols] = get(j,k); 
    }
  }
  
  delete A;

  columns = newcols;
  rows = newrows;

  A = newA;
}

double matrix::partzero() {
  int j=0;
  for(int i=0;i<rows*columns;i++) {
    if(A[i]==0) {
      j++;
    }
  }
  return (double) j/(rows*columns);
}
